/*
MP3 Extractor v1.0
Chapin Information Services, Inc.
info-svc.com
10/16/2006

Description:  Utility quickly restores fragmented MP3 files.  This version looks for fixed-length fragments in MP3 files only.  No installation required.

Concept:  Media encapsulation can be removed without knowing the exact format.

Distribution:  Original files from .info-svc.com.  Private use only.  Free.  Not for resale.  Includes source code.

Thanks:
http://mpgedit.org/mpgedit/mpeg_format/mpeghdr.htm
http://www.id3.org/id3v2.4.0-structure.txt
http://msdn.microsoft.com/vstudio/express/visualc/

Known Bugs:
  None

History:
  10/16/06 Build 1  Added VBR handling.  Added checks for locked files.  Fixed off-by-one bugs in array functions.  Optimized main loop.
  10/02/06 Build 0
*/


#include "stdafx.h"


int SearchByteArray(unsigned char LookIn[], int LookInLength, unsigned char LookFor[], int LookForLength, int Start = 0);

int MP3FrameSize(unsigned char FrameSizeInfo, int& samplerate);

bool CompareByteArrays(unsigned char Array1[], int Array1Length, int Array1Start, unsigned char Array2[], int Array2Length);

        
int _tmain(int argc, _TCHAR* argv[])
{

    const int InitialBufferSize = 20000;

    unsigned char ID3v2Begin[3] = {'I', 'D', '3'};
    unsigned char MP3FrameBeginCRC[2] = {0xFF, 0xFA};
    unsigned char MP3FrameBeginNoCRC[2] = {0xFF, 0xFB};

    unsigned char *MP3FrameBegin;  //synch first two bytes
    unsigned char MP3FrameBegin2;  //synch fourth byte

    int InFile = 0;
    int OutFile = 0;
    FILE *FileInStream;
    FILE *FileOutStream;

    unsigned char* FileBuffer;

    int samplerate = 0;

    int CurrPos = 0;
    int TempPos = 0;
    int BeginPos = 0;
    int FileSize = 0;
    int FrameSize = 0;
    int Counter1 = 0;
    unsigned int Counter2 = 0;
    int Counter3 = 0;
    unsigned char FirstDWord[4];
    char Path[251];
    char *NewPath;
    bool ExtraDataFound = false;
    int ExtraInfoSize = 0;
    int ExtraInfoInterval = 0;
    int EndOfFile1 = 0;
    int EndOfFile2 = 0;

    if(argc > 1 && !((char)argv[1][0] == '/' && (char)argv[1][1] == '?')){
        for(Counter1 = 1; Counter1 < argc; Counter1++){
            for(Counter2 = 0; Counter2 <= _tcslen(argv[Counter1]) && Counter3 < 251; Counter2++){
                Path[Counter3] = (char)argv[Counter1][Counter2];
                if(Path[Counter3] == 0){Path[Counter3] = 32;}
                Counter3++;
            }
        }
        Path[Counter3 - 1] = 0;
    }else{
        _cputs("MP3 Extractor v1.0.001\n\nChapin Information Services, Inc.\ninfo-svc.com\n\nmp3extract filename\n\n");
        exit(1);
    }

    NewPath = new char[Counter3 + 4];
    strncpy_s(NewPath, Counter3, Path, Counter3);
    strcat_s(NewPath, Counter3 + 4, ".mp3");

    _sopen_s(&InFile, Path, 0, _SH_DENYNO, _S_IREAD);
    if(InFile < 0){
        _cputs("Source File Could Not Be Found\n");
        exit(3);
    }
    FileSize = _filelength(InFile);
    _close(InFile);
    _sopen_s(&OutFile, NewPath, 0, _SH_DENYNO, _S_IREAD);
    if(OutFile >= 0){
        _cputs("Destination File Already Exists, Aborting!\n");
        _close(OutFile);
        exit(2);
    }

    FileBuffer = new unsigned char[InitialBufferSize];

    FileInStream = _fsopen(Path, "rb", _SH_DENYWR);
    if(!FileInStream){
        _cputs("Unable to open source file, it may be open in another program.\n");
        exit(9);
    }
    fread(FileBuffer, 1, InitialBufferSize, FileInStream);

    TempPos = SearchByteArray(FileBuffer, InitialBufferSize, MP3FrameBeginCRC, 2);
    CurrPos = SearchByteArray(FileBuffer, InitialBufferSize, MP3FrameBeginNoCRC, 2);

    if(TempPos == 0){
        if(CurrPos == 0){
            _cputs("Could not find any MP3 frames in source file.\n");
            exit(4);
        }else{
            TempPos = CurrPos;
            MP3FrameBegin = MP3FrameBeginNoCRC;
        }
    }else{
        MP3FrameBegin = MP3FrameBeginCRC;
    }

    MP3FrameBegin2 = FileBuffer[TempPos + 3];

    CurrPos = SearchByteArray(FileBuffer, InitialBufferSize, ID3v2Begin, 3);

    if(CurrPos > 0 && CurrPos < TempPos){
        BeginPos = CurrPos;
    }else{
        BeginPos = TempPos;
    }

    //Determine FirstDWord of extra info

    FrameSize = MP3FrameSize(FileBuffer[TempPos + 2], samplerate);

    do{

        CurrPos = SearchByteArray(FileBuffer, InitialBufferSize, MP3FrameBegin, 2, TempPos + 2);
        if(CurrPos == 0){
            _cputs("Could not find any non-MP3 data after first frame in source file.\n");
            exit(5);
        }else if(FileBuffer[CurrPos + 3] != MP3FrameBegin2){
            TempPos = CurrPos;
        }else if(CurrPos > (TempPos + FrameSize)){
            ExtraDataFound = true;
            ExtraInfoSize = CurrPos - (TempPos + FrameSize);
        }else{
            TempPos = CurrPos;
            FrameSize = MP3FrameSize(FileBuffer[TempPos + 2], samplerate);
        }

    }while(!ExtraDataFound);

    TempPos = BeginPos - ExtraInfoSize;

    FirstDWord[0] = FileBuffer[TempPos];
    FirstDWord[1] = FileBuffer[TempPos + 1];
    FirstDWord[2] = FileBuffer[TempPos + 2];
    FirstDWord[3] = FileBuffer[TempPos + 3];

    //Determine ExtraInfoInterval

    CurrPos = TempPos;

    do{

        CurrPos = SearchByteArray(FileBuffer, InitialBufferSize, FirstDWord, 4, CurrPos + 4);

        if(CurrPos > 0){
            ExtraInfoInterval = CurrPos - TempPos;
        }else{
            _cputs("Did not find a pattern in the extra info of the source file.\n");
            exit(6);
        }

    //Check for 3rd apperance of extra info
    }while(!CompareByteArrays(FileBuffer, InitialBufferSize, CurrPos + ExtraInfoInterval, FirstDWord, 4));

    FrameSize = ExtraInfoInterval - ExtraInfoSize;
    EndOfFile1 = FileSize - FrameSize;
    EndOfFile2 = FileSize - ExtraInfoSize;

    delete [] FileBuffer;
    FileBuffer = new unsigned char[FrameSize];

    fseek(FileInStream, BeginPos, SEEK_SET);
    FileOutStream = _fsopen(NewPath, "wb", _SH_DENYWR);

    do{

        fread(FileBuffer, 1, FrameSize, FileInStream);
        fwrite(FileBuffer, 1, FrameSize, FileOutStream);
        if(ftell(FileInStream) <= EndOfFile2){
            fread(FileBuffer, 1, ExtraInfoSize, FileInStream);
        }

    }while(CompareByteArrays(FileBuffer, FrameSize, 0, FirstDWord, 4) && ftell(FileInStream) <= EndOfFile1);

    fclose(FileOutStream);
    fclose(FileInStream);

    return 0;
}

int MP3FrameSize(unsigned char FrameSizeInfo, int& samplerate){

    int bitrate = 0;

    switch (FrameSizeInfo & 0xF0){
        case 0x10:
            bitrate = 32000;
            break;
        case 0x20:
            bitrate = 40000;
            break;
        case 0x30:
            bitrate = 48000;
            break;
        case 0x40:
            bitrate = 56000;
            break;
        case 0x50:
            bitrate = 64000;
            break;
        case 0x60:
            bitrate = 80000;
            break;
        case 0x70:
            bitrate = 96000;
            break;
        case 0x80:
            bitrate = 112000;
            break;
        case 0x90:
            bitrate = 128000;
            break;
        case 0xA0:
            bitrate = 160000;
            break;
        case 0xB0:
            bitrate = 192000;
            break;
        case 0xC0:
            bitrate = 224000;
            break;
        case 0xD0:
            bitrate = 256000;
            break;
        case 0xE0:
            bitrate = 320000;
            break;
        default:
            if(samplerate == 0){
                _cputs("Invalid bitrate in first MP3 frame.\n");
                exit(7);
            }else{
                _cputs("Invalid bitrate after first MP3 frame.\n");
                exit(11);
            }
            break;
    }
    
    if(samplerate == 0){
        switch(FrameSizeInfo & 0xC){
            case 0x0:
                samplerate = 44100;
                break;
            case 0x4:
                samplerate = 48000;
                break;
            case 0x8:
                samplerate = 32000;
                break;
            default:
                _cputs("Invalid sample rate in first MP3 frame.\n");
                exit(8);
                break;
        }
    }

    return((144 * bitrate / samplerate) + ((FrameSizeInfo & 0x2) >> 1));

}

int SearchByteArray(unsigned char LookIn[], int LookInLength, unsigned char LookFor[], int LookForLength, int Start){

    int Counter = 0;
    int SubCounter = 0;
    int Found = 0;

    if(LookForLength > 0 && LookForLength <= LookInLength && Start >= 0 && Start <= LookInLength - LookForLength){

        for(Counter = Start; Counter <= LookInLength - LookForLength; Counter++){

            if(LookIn[Counter] == LookFor[0]){

                Found = Counter;

                for(SubCounter = 1; SubCounter < LookForLength && Found > 0; SubCounter++){

                    if(LookIn[Counter + SubCounter] != LookFor[SubCounter]){
                        Found = 0;
                    }
                }
                if(Found > 0){return Found;}
            }
        }
    }
    return Found;
}

bool CompareByteArrays(unsigned char Array1[], int Array1Length, int Array1Start, unsigned char Array2[], int Array2Length){

    int Counter = 0;

    if(Array2Length >= 0 && Array2Length <= Array1Length && Array1Start >= 0 && Array1Start <= Array1Length - Array2Length){

        for(Counter = 0; Counter < Array2Length; Counter++){

            if(Array1[Counter + Array1Start] != Array2[Counter]){
                return(false);
            }
        }
    }else{
        return(false);
    }
    return(true);
}